from __future__ import print_function, division, absolute_import
from future import standard_library

standard_library.install_aliases()
from birdseye.server import main

if __name__ == '__main__':
    main()
