<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>{% block title %}Bird's Eye{% endblock %}</title>
    <link rel="stylesheet" href="/static/css/main.css">
    <link rel="stylesheet" href="/static/css/bootstrap.min.css">
    <script src="/static/js/libs/jquery-3.2.1.js"></script>
    <script src="/static/js/libs/bootstrap.js"></script>
    {% block head %}{% endblock %}
</head>
<body>
<div class="container">
    <a href="/">
        <h1>
            Bird's Eye
        </h1>
    </a>
    <hr>
    {% block body %}{% endblock %}
</div>
</body>
</html>